"""Test the Location class."""

import uk_politics
import uk_politics.location


def test_location_str():
    """Test the str(location) property."""
    count = uk_politics.elections.COUNTS[0]
    location = count.location
    for property_name in ["id", "constituency",
                          "county", "country", "region", "electorate"]:
        assert property_name + ":" in str(location)


def test_comparisons():
    """Test the >= and <= comparisons.

    These comparisons are only supposed to be performed
    where one side is a fully populated Location object (from the data)
    and the other is a filter created by the user.
    """
    location_one = uk_politics.Location(constituency="Home")
    location_two = uk_politics.Location(constituency="Away")
    assert location_one <= location_one
    assert location_one >= location_one
    assert not location_one <= location_two
    assert not location_one >= location_two


def test_location_from_string():
    """Turn a fully-populated Location object into a string and back."""
    location = uk_politics.elections.COUNTS[0].location
    as_string = repr(location)
    from_string = uk_politics.location.from_raw_string(as_string)
    assert location == from_string


def test_location_from_string_with_wildcards():
    """Turn a fully-populated Location object into a string and back."""
    location = uk_politics.location.from_raw_string("*|*|*|*|*|0")
    assert location.constituency is None
    assert location.electorate == 0



"""Tests for the data_tables module."""
import pytest
import uk_politics.exceptions
import uk_politics.data_tables


def test_file_not_found_exception():
    """Checking for file not found expection when loading data table."""
    with pytest.raises(uk_politics.exceptions.DataFileMissingOrUnreadable):
        uk_politics.data_tables.test_data_file(".csv")



"""Test the contents of the results submodule."""
from datetime import datetime
import uk_politics


def test_all_votes():
    """Check at least one vote record."""
    assert len(uk_politics.elections.COUNTS) > 0


def test_filter_by_constituency():
    """Check filtering by Constituency.

    Check at least one result,
    check all results in right location.
    """
    location_filter = uk_politics.Location(constituency="BUCKINGHAM")
    buckingham_results = [vote for vote in uk_politics.elections.COUNTS
                          if location_filter >= vote.location]

    # Check these results are all for the right constituency
    print(buckingham_results[0])
    for vote in buckingham_results:
        assert vote.location.constituency == "BUCKINGHAM"
    assert len(buckingham_results) > 0


def test_filter_by_region():
    """Check filtering by Region.

    Check at least one result,
    check all results in right location.
    """
    location_filter = uk_politics.Location(region="Wales")
    wales_results = [vote for vote in uk_politics.elections.COUNTS
                     if location_filter >= vote.location]

    print(wales_results[0])
    for vote in wales_results:
        assert vote.location.region == "Wales"
    assert len(wales_results) > 0


def test_filter_by_date():
    """Check filtering by Date.

    Check at least one result,
    check all results are for given date.
    """
    date_filter = datetime(1997, 5, 1).date()
    election_results = [vote_total for vote_total in uk_politics.elections.COUNTS
                        if date_filter == vote_total.date]

    for vote in election_results:
        assert vote.date == date_filter
    assert len(election_results) > 0


def test_speaker():
    """Test that precisely one result each year is ascribed to Speaker."""
    for date in uk_politics.elections.DATES:
        election_results = [vote_total
                            for vote_total in uk_politics.elections.COUNTS
                            if date == vote_total.date]
        speaker_results = [vote_total for vote_total in election_results
                           if vote_total.party == "Speaker"]
        assert len(speaker_results) == 1


def test_electorate():
    """Test that the electorate count in a constituency makes sense."""
    election_2019 = datetime(2019, 12, 12).date()
    counts = [count for count in uk_politics.elections.COUNTS
              if uk_politics.Location(constituency="DUNDEE EAST") >= count.location
              if count.date == election_2019]
    electorate_first = counts[0].location.electorate
    for count in counts:
        assert count.location.electorate == electorate_first

    vote_turnout = sum([count.votes for count in counts]) / electorate_first
    assert abs(vote_turnout - 0.68) < 0.01



"""Test the contents of uk_politics.elections."""

import datetime
import uk_politics


def test_ran():
    """Sixteen parties ran that year, including "Speaker"."""
    assert len(uk_politics.elections.ran(datetime.date(2019, 12, 12))) == 16


def test_seats():
    """Should be 650."""
    assert len(uk_politics.elections.seats()) == 650

def test_seats_is_list():
    """Should return a list of tuples."""
    assert len(uk_politics.elections.seats()[0]) == 2



r"""A container for (possibly wildcard) location data."""

import dataclasses
from typing import Any, List, Optional


@dataclasses.dataclass()
class Location:
    """Location specification.

    Leave options as None to act as wildcards.
    The ons_id term in particular in inconsistent between elections
    (or at least elections up until the standardisation of ONS ids)
    so is best left as None.
    Comparisons are made as though lower case.

    When comparing Locations:
        A >= B if A is more general than, or equal to, B.
    Note that this comparison only properly works when B
    contains no wildcards. Use of "A > B" is discouraged.

    When inspecting str(Location) the None entries
    are represented as "*" as befits a wildcard.

    Attributes:
        constituency: str
        county: str
        region: str
        country: str
        ons_id: str
        electorate: int
    """

    constituency: Optional[str] = None
    county: Optional[str] = None
    region: Optional[str] = None
    country: Optional[str] = None
    ons_id: Optional[str] = None
    electorate: Optional[int] = None

    def __init__(self,
                 ons_id: Optional[str] = None,
                 constituency: Optional[str] = None,
                 county: Optional[str] = None,
                 region: Optional[str] = None,
                 country: Optional[str] = None,
                 electorate: Optional[int] = None):
        """Create a location filter.

        There is no difference between a Location object
        and a location filter; but users are expected to
        only ever create filters. Location objects are populated
        from the data files at import time.
        """
        self.ons_id = ons_id
        self.constituency = constituency
        self.county = county
        self.region = region
        self.country = country
        self.electorate = electorate

    def _contains_loc_as_raw_string(self, raw_string: str) -> bool:
        """Compare self against a raw string.

        None is treated as a wildcard.
        Ignores the `electorate` attribute.
        """
        string_properties = raw_string.split("|")
        for index, prop in enumerate(self._string_properties_list):
            if prop is not None:
                if prop.lower() != string_properties[index].lower():
                    return False

        return True

    def __ge__(self, other: "Location") -> bool:
        """Compare against another Location object.

        self >= other if self is more general.
        """
        return self._contains_loc_as_raw_string(repr(other))

    def __le__(self, other: "Location") -> bool:
        """Compare against another Location object.

        self <= other if other is more general.
        """
        return other.__ge__(self)

    @property
    def _string_properties_list(self) -> List[Optional[str]]:
        """Those properties that take string values.

        These are the properties used for comparison.
        """
        return [self.ons_id,
                self.constituency,
                self.county,
                self.region,
                self.country]

    def __str__(self) -> str:
        """Location data in human-readable format."""
        return (f"id:{_wildcard_if_none(self.ons_id)} "
                f"constituency:{_wildcard_if_none(self.constituency)} "
                f"county:{_wildcard_if_none(self.county)} "
                f"region:{_wildcard_if_none(self.region)} "
                f"country:{_wildcard_if_none(self.country)} "
                f"electorate:{_wildcard_if_none(self.electorate)}")

    def __repr__(self) -> str:
        """Coerce to |-separated string."""
        return "|".join(list(map(_wildcard_if_none, self._string_properties_list))
                        + [_wildcard_if_none(self.electorate)])

    def __eq__(self, other: object) -> bool:
        """Check equality by checking string representation.

        Note that this will not be stable between years,
        since the electorate will be of different sizes.
        You may want to instead compare ons_id or name.
        """
        return repr(self) == repr(other)

    def __hash__(self) -> int:
        """Hash based on repr."""
        return hash(repr(self))


def _wildcard_if_none(potentially_none: Any) -> str:
    """Turn None into *, otherwise returns str form of argument."""
    if potentially_none is None:
        return "*"
    return str(potentially_none)


def from_raw_string(raw: str) -> Location:
    """Turn a |-separated string into a Location object."""
    def wildcard_to_nonetype(string):
        if string == "*":
            return None
        return string
    split: List[Optional[str]] = list(
        map(wildcard_to_nonetype, raw.split("|"))
    )
    if split[5] is None:
        electorate = None
    else:
        electorate = int(split[5])
    return Location(split[0], split[1], split[2],
                    split[3], split[4], electorate)

"""MIT License

Copyright (c) 2021 Stonehaven

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


r"""Election vote counts.

Election results are stored as a VoteTotal object;
one for each party / location / date combination.
Speakers are assigned to the party "Speaker".
"""
import dataclasses
import datetime
from typing import Dict, List, Optional, Tuple

from uk_politics import exceptions

from . import data_tables
from . import location
from . import names


@dataclasses.dataclass
class VoteTotal:
    """Vote total for a single party / location / date combination.

    Attributes:
        location (uk_politics.Location): Single constituency.
        date (datetime.date): Date on which the election happened.
        party (str): Official, current, name of the party.
        party_contemporary_styling (str): Party as recorded in the data.
        votes (int): Vote count for this party in this place on this day.

    The property `.party` is always the modern official name for the party,
    and `.party_contemporary_styling` is the
    short form used in the election data.
    For example some elections just write "PC" instead of
    "Plaid Cymru - the Party of Wales", so we record both.
    """

    location: location.Location
    date: datetime.date
    party: str
    party_contemporary_styling: str
    votes: int


def _vote_line_to_object(line: List[str]) -> VoteTotal:
    """Convert a line in a tsv file to a VoteTotal object."""
    count_location = location.from_raw_string(line[0])
    party_contemporary_styling = line[2]
    if party_contemporary_styling == "PC/SNP":
        # These were lumped together in some elections
        if count_location.country == "Scotland":
            party_contemporary_styling = "SNP"
        elif count_location.country == "Wales":
            party_contemporary_styling = "PC"
    elif party_contemporary_styling == "Green":
        # Green party is officially distinct between these regions
        # Use names.short to turn these all back into "Green Party"
        if count_location.country == "Scotland":
            party_contemporary_styling = "Scottish Green Party"
        elif count_location.country == "Northern Ireland":
            party_contemporary_styling = "Green Party Northern Ireland"
        else:
            party_contemporary_styling = "Green Party of England and Wales"

    party_official_name = names.official(
        party_contemporary_styling, warn_on_fuzzy_match=True)

    if party_official_name is None:
        raise exceptions.DataLineUnreadable(
            "\t".join(line), "party name could not be read")

    return VoteTotal(location=count_location,
                     date=datetime.datetime.strptime(
                         line[1], "%Y-%m-%d").date(),
                     party=party_official_name,
                     party_contemporary_styling=party_contemporary_styling,
                     votes=int(line[3]))


_ALL_VOTES_DF = data_tables.read_csv_to_dataframe("GE_results.tsv", "\t")

COUNTS: List[VoteTotal] = [_vote_line_to_object(vote[1])
                           for vote in _ALL_VOTES_DF.iterrows()]
"""A list of every party / date / location vote count."""

PARTIES_ORIGINAL_NAMES: List[str] = sorted(
    list({vote_count.party for vote_count in COUNTS}))
"""A list of every party name as styled in the original data."""

DATES: List[datetime.date] = sorted(
    list({vote_count.date for vote_count in COUNTS}))
"""The dates of the elections for which we have data."""

MOST_RECENT_ELECTION_DATE: datetime.date = max(DATES)
"""The date of the most recent election in the data."""


def seats(date: datetime.date = MOST_RECENT_ELECTION_DATE,
          location_filter: Optional[location.Location] = None
          ) -> List[Tuple[location.Location, str]]:
    """List the constituencies along with the name of the party that won.

    The Speaker is treated as their own party.
    See politics.elections.DATES for a list of dates.

    Args:
        date (datetime.date, optional):
            The date of the election to inspect.
            Defaults to MOST_RECENT_ELECTION_DATE.

    Returns:
        A list of tuples (List[Tuple[location.Location, str]]):
            Location
            Winning party
    """
    if location_filter is None:
        location_filter = location.Location()
    counts_for_this_date = [count for count in COUNTS
                            if location_filter >= count.location
                            if count.date == date]
    locations = {count.location for count in counts_for_this_date}
    constituency_winners: Dict[location.Location, str] = {}
    for loc in locations:
        party_counts = {count.party: count.votes
                        for count in counts_for_this_date
                        if count.location == loc}
        winner = max(party_counts.keys(), key=lambda x,
                     counter=party_counts: counter[x])
        constituency_winners[loc] = winner
    return list(constituency_winners.items())


def ran(
    date: datetime.date = MOST_RECENT_ELECTION_DATE,
    location_filter: Optional[location.Location] = None,
) -> List[str]:
    """Parties that ran on the given date in the given location.

    Leave location_filter as None to cover all locations.

    Args:
        date (datetime.date, optional):
            Date of election to inspect.
            Defaults to MOST_RECENT_ELECTION_DATE.
        location_filter (location.Location, optional):
            Only show parties that ran in locations within location_filter.
            Defaults to None (most general).

    Returns:
        List of parties (List[str])
    """
    if location_filter is None:
        parties = {count.party for count in COUNTS
                   if date == count.date}
    else:
        parties = {count.party for count in COUNTS
                   if location_filter >= count.location
                   if date == count.date}
    return sorted(list(parties))

"""MIT License

Copyright (c) 2021 Stonehaven

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


"""Parties have official names, short names, and nicknames.

This submodule attempts to link these different types of name.
The short name is the colloquial name as used by
https://www.parliament.uk/about/mps-and-lords/members/parties/ .
Example:
 - official name: Conservative and Unionist Party
 - short name: Conservative Party
 - nickname: Tory

The nicknames come from our own survey data.

The "Green Party" is not one party but a collection of parties,
with distinct chapters in Scotland, Northern Ireland,
and England & Wales. The election data, when presented to the user,
distinguishes between these parties with their official names
(which contain their country identifier), but all three official
names are shortened to Green Party.

The Co-operative Party is in alliance with the Labour party,
and all their seats are recorded as Labour seats in the data.

The Speaker is regarded as a party of one member.
"""

import logging
import functools
from typing import Optional

import pandas as pd
import fuzzywuzzy.process

from . import data_tables
from . import exceptions

_logger = logging.getLogger("uk-politics")

@functools.lru_cache
def official(nickname: Optional[str],
             allow_fuzzy_match=True,
             warn_on_fuzzy_match=True,
             exception_on_null_value=False) -> Optional[str]:
    """Return the official name of a party from a given nickname.

    The function uses fuzzy match (Levenstein distance, from fuzzywuzzy)
    to find the closest match by default;
    set `allow_fuzzy_match=False` to turn this off.
    These renames will appear as warnings.

    This function is cached to avoid running extra fuzzy matches,
    this also means that warnings will only appear the first time
    a given renaming takes place. By default nicknames that `pandas`
    considers to be a null value are passed through as None.

    Args:
        nickname (str): [description]
        allow_fuzzy_match (bool, optional): [description]. Defaults to True.
        warn_on_fuzzy_match (bool, optional): [description]. Defaults to True.

    Raises:
        exceptions.PartyNicknameEmpty: [description]
        exceptions.PartyNameNotFound: [description]

    Returns:
        official_name (str): The official name for the party
        that best matches our nicknames on record.
        For example:
            official_name("tories") -> "Conservative and Unionist Party"
    """
    if pd.isna(nickname):
        if exception_on_null_value:
            raise exceptions.PartyNicknameEmpty()
        return None

    assert nickname is not None # for typing

    nickname = nickname.lower().strip()
    if len(nickname) == 0:
        raise exceptions.PartyNicknameEmpty()

    if nickname in data_tables.PARTY_NICKNAMES:
        # In our database of nicknames
        return data_tables.PARTY_NICKNAMES[nickname]

    if allow_fuzzy_match:
        #types: ignore
        fuzzy_matched = fuzzywuzzy.process.extractOne(
            nickname, list(data_tables.PARTY_NICKNAMES.keys()))[0]
        proper_name = data_tables.PARTY_NICKNAMES[fuzzy_matched]
        if warn_on_fuzzy_match:
            _logger.warning("Renaming '%s' -> '%s'", nickname, proper_name)
        return proper_name

    raise exceptions.PartyNameNotFound(nickname)


def short(official_name: Optional[str]) -> Optional[str]:
    """Short name we use instead of the given official party name.

    Args:
        official_name (str): [description]

    Raises:
        exceptions.PartyNameNotFound: [description]

    Returns:
        shot_name (str): The short name for a given official party name.
        For example:
            short_name("Conservative and Unionist Party") -> "Conservative"
    """
    if official_name is None:
        return None
    official_name = official_name.lower()
    if official_name in data_tables.PARTY_SHORTNAMES:
        return data_tables.PARTY_SHORTNAMES[official_name]

    raise exceptions.PartyNameNotFound(official_name)

"""MIT License

Copyright (c) 2021 Stonehaven

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


"""Test the party module functionality."""

import warnings
import pytest
import uk_politics
import uk_politics.data_tables
import uk_politics.exceptions

CONSERVATIVES_NICKNAMES = ["tory", "Tory",
                           "torie", "Torie",
                           "torys", "Torys",
                           "tories", "Tories"]


def test_find_tories():
    """Understand variations on "tory" as "Conservative and Unionist Party"."""
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", module="uk_politics")
        for nickname in CONSERVATIVES_NICKNAMES:
            print(nickname)
            assert uk_politics.names.official(
                nickname) == "Conservative and Unionist Party"

        assert uk_politics.names.short(
            "Conservative and Unionist Party") == "Conservative Party"


def test_do_not_find_tories_without_fuzzy():
    """Without fuzzy matching we raise an exception."""
    for nickname in [nickname
                     for nickname in CONSERVATIVES_NICKNAMES
                     if nickname.lower() not in
                     uk_politics.data_tables.PARTY_NICKNAMES]:
        print(nickname)
        with pytest.raises(uk_politics.exceptions.PartyNameNotFound):
            uk_politics.names.official(nickname, allow_fuzzy_match=False)


EXTRANEOUS_PARTIES = {"England":
                      ["Scottish National Party", "Sinn Fein",
                       "Democratic Unionist Party", "Plaid Cymru"],
                      "Scotland":
                      ["Sinn Fein", "Democratic Unionist Party",
                       "Plaid Cymru"],
                      "Northern Ireland":
                      ["Scottish National Party",
                       "Plaid Cymru"],
                      "Wales":
                      ["Scottish National Party", "Sinn Fein",
                       "Democratic Unionist Party"]
                      }


def test_current_parties_for_edge_cases():
    """Listed parties should not appear in the given country."""
    for country, party_list in EXTRANEOUS_PARTIES.items():
        official_name_party_list = list(
            map(uk_politics.names.official, party_list))
        print(country, official_name_party_list)

        parties_in_this_country = uk_politics.elections.ran(
            location_filter=uk_politics.Location(country=country))
        print(parties_in_this_country)

        assert set(parties_in_this_country).isdisjoint(
            official_name_party_list)


today_party_list = ["Conservatives",
                    "Democratic Unionist Party",
                    "GPEW",
                    "Green Party of Scotland",
                    "Green Party of Northern Ireland",
                    "Labour",
                    "Liberal Democrats",
                    "Other",
                    "Plaid Cymru",
                    "Reform UK (formerly Brexit party)",
                    "Scottish National Party",
                    "Sinn Fein"]


def test_current_parties():
    """Check that all parties listed ran in 2019."""
    for party in today_party_list:
        official_name = uk_politics.names.official(party)
        assert official_name in uk_politics.elections.ran()


def test_green_party():
    """The Green Party is the short name for multiple parties."""
    assert uk_politics.names.official("green party") == "Green Party"


def test_empty_name():
    """Test passing empty nickname raises the right error."""
    with pytest.raises(uk_politics.exceptions.PartyNicknameEmpty):
        uk_politics.names.official("")


def test_null_name_value():
    """Test null value for name data."""
    with pytest.raises(uk_politics.exceptions.PartyNicknameEmpty):
        uk_politics.names.official(None, exception_on_null_value=True)

    assert uk_politics.names.official(None) is None


def test_short_null_name_value():
    """Test null value for name data."""
    assert uk_politics.names.short(None) is None


def test_short_name_invalid_value():
    """Test invalid party name."""
    with pytest.raises(uk_politics.exceptions.PartyNameNotFound):
        uk_politics.names.short("None")



"""Test the party_color functionality."""

import re
import pytest
import uk_politics
import uk_politics.data_tables
import uk_politics.exceptions


def test_example_given():
    """party_color("Brexit Party") -> "#12B6CF" is in the docstring."""
    assert uk_politics.color("Reform UK") == "#12B6CF"


def test_invalid_party_name():
    """Passing an invalid party should raise an exception."""
    with pytest.raises(uk_politics.exceptions.NoColorForThisParty):
        uk_politics.colors.by_official_party_name("")


def test_hex_code():
    """All the values should be hex codes."""
    for color in uk_politics.data_tables.PARTY_COLORS.values():
        assert re.match("^#[0-9A-F]{6}$", color) is not None


def test_no_vote():
    """Party "No vote" is a party from the point of view of having a color."""
    assert uk_politics.color("No vote") == "#333333"


def test_green_party():
    """The green party has warnings regarding color choice."""
    with pytest.warns(uk_politics.exceptions.UKPoliticsWarning):
        uk_politics.color("Green")



"""A namespace for exception handling."""


class UKPoliticsException(Exception):
    """Base exception class for the politics module."""


class UKPoliticsWarning(Warning):
    """Base warning class for the politics module."""


class DataLineUnreadable(UKPoliticsException):
    """A line in the data isn't well-formed."""

    def __init__(self, offending_line: str, additional_context):
        """Raise exception citing line and issue."""
        super().__init__(
            (f"The line '{offending_line}' could not"
             f" be parsed because {additional_context}"))


class DataFileMissingOrUnreadable(UKPoliticsException):
    """Exception for missing or unreadable data file."""

    def __init__(self, short_name: str):
        """Raise exception with name of offending file."""
        super().__init__(
            f"Data file {short_name} missing or cannot be read."
        )


class PartyNameNotFound(UKPoliticsException):
    """Could not find a party of the given name."""

    def __init__(self, needle: str):
        """Raise exception announcing problem name."""
        super().__init__(f"No party with name '{needle}' was found.")


class PartyNicknameEmpty(UKPoliticsException):
    """No party can have an empty or null nickname."""

    def __init__(self):
        """Raise exception announcing empty nickname."""
        super().__init__("The search string is not allowed"
                         " to be empty or null.")


class NoColorForThisParty(UKPoliticsException):
    """Could not find a color for the given party."""

    def __init__(self, needle: str):
        """Raise exception announcing problem name."""
        super().__init__(f"No color was found for party '{needle}'.")

"""MIT License

Copyright (c) 2021 Stonehaven

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


r"""UK Politics.

Covers the names and colors of the UK political parties,
as well as election vote counts for the major parties in
every election and constituency since 1997.
This is based on data released by the House of Commons,
see GE_results.tsv.licence for details.

Example uses:
    uk_politics.elections.seats(datetime.date(2019,12,12)) ->
        list constituencies and winning parties
    uk_politics.color("Labour Party") -> "#E4003B"
    uk_politics.find_party("Tories") -> "Conservative and Unionist Party"
    uk_politics.find_party("Tories", return_short_name=True) -> "Conservative Party"

Licencing summary:
    This module contains Parliamentary information
    licensed under the Open Parliament Licence v3.0.
    The data files (in the data directory) contain
    licencing information alongside each file.

Submodules:
    uk_politics.elections holds vote count data for each election
    uk_politics.names holds functions for official-, short-, and nick-names
    uk_politics.colors maps party names to colors

Hidden submodules:
    uk_politics.data_tables handles most of the data extraction
    uk_politics.location
        holds a custom Location class (useful for filtering election data)
        exposed as politics.Location
    uk_politics.exceptions holds exceptions

Spelling:
    In the interest of international consistency we will use EN_US spellings.

"""

from typing import Optional
from . import colors
from . import elections
from . import names

# Bring the following elements forward in the namespace
from .location import Location


def color(party_name_or_nickname: Optional[str]) -> Optional[str]:
    """Color (hex code) for a party based on a name or nickname.

    For finer control see politics.names and politics.colors.
    Returns the current color for a party.

    Args:
        party_name_or_nickname (str)

    Returns:
        color (str): The hex code (including "#") assigned to that party.
        For example:
            color("Tory") -> '#0087DC'
    """
    official_name = names.official(party_name_or_nickname)
    return colors.by_official_party_name(official_name)


def find_party(nickname: Optional[str], return_short_name: bool = False, allow_fuzzy: bool = True) -> Optional[str]:
    """Find a party based on nickname.

    By default this allows fuzzy matching,
    but warns the user when it is applied.
    For finer control see `politics.names.official` and
    `politics.names.short`.

    Args:
        nickname (str): unofficial name for party you want to search for
        return_short_name (bool, optional):
            Return short name, rather than the sometimes unwieldy
            official name. Defaults to False.
        allow_fuzzy (bool, optional):
            Allow fuzzy searching to try and find the correct party.
            Defaults to True.

    Returns:
        name (str): The official name of the party best matching the nickname.
    """
    official_name = names.official(nickname, allow_fuzzy_match=allow_fuzzy)

    if return_short_name:
        return names.short(official_name)

    return official_name


"""MIT License

Copyright (c) 2021 Stonehaven

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


"""Test the functions exposed at the top level of the module.

This isn't a full test of each method's capabilities,
just checking that the method is exposed at the top level namespace.
"""

import warnings
import pytest
import uk_politics
import uk_politics.exceptions


def test_color():
    """Check that uk_politics.color works."""
    assert uk_politics.color("Liberal Democrats") == "#FAA61A"


def test_location():
    """Check that uk_politics.Location works.

    Just creation and comparison here.
    """
    wales = uk_politics.Location(country="Wales")
    west_glamorgan = uk_politics.elections.COUNTS[0].location
    assert wales >= west_glamorgan
    assert not west_glamorgan >= wales


def test_find_party():
    """Check that uk_politics.find_party works."""
    assert uk_politics.find_party(
        "Tory", return_short_name=True) == "Conservative Party"


def test_find_empty():
    """Passing an empty string should raise assertion error."""
    with pytest.raises(uk_politics.exceptions.PartyNicknameEmpty):
        assert uk_politics.find_party("") == ""


def test_scottish_labour():
    """Test variations on Scottish Labour return Labour not SNP.

    The name is a bit too close to two different parties.
    """
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", module="uk_politics")
        assert uk_politics.find_party("Scottish Labour") == "Labour Party"
        assert uk_politics.find_party("Scottish labour") == "Labour Party"


def test_rename_gives_warning(caplog):
    """Test that a bad name prompts a rename warning."""
    uk_politics.find_party("labuor")
    print(caplog.records)
    expected = "Renaming 'labuor' -> 'Labour Party'"
    assert expected in caplog.text



"""Functions for importing data."""
import os
import csv
from typing import Dict, Tuple
import pandas as pd

from . import exceptions


def data_path(filename: str) -> str:
    """Get full path to a file in the data folder.

    Args:
        filename (str): The data file's name.

    Returns:
        path (str): The full path to the file.
    """
    return os.path.join(os.path.dirname(__file__), "data", filename)


def test_data_file(short_name: str) -> None:
    """Raise exception if data file unavailable.

    Args:
        short_name (str): file name.

    Raises:
        exceptions.DataFileMissingOrUnreadable

    Returns:
        None
    """
    path = data_path(short_name)
    if not os.path.isfile(path):
        raise exceptions.DataFileMissingOrUnreadable(short_name)


def read_csv_to_dict(short_name: str) -> Dict[str, str]:
    """Load csv into memory.

    Args:
        short_name (str): file name.

    Returns:
        Dict[str, str]: A dictionary of first column -> second column
    """
    path = data_path(short_name)
    test_data_file(short_name)
    with open(path, "r", encoding="utf-8-sig") as file:
        csv_reader = csv.reader(file, delimiter=",")
        mapping = {rows[0]: rows[1] for rows in csv_reader}
    return mapping


def read_nicknames_file(
    short_name: str
) -> Tuple[Dict[str, str], Dict[str, str]]:
    """Read a nicknames csv file into two dictionaries.

    The file is in the format:
        official,short,*other_nicknames

    Args:
        short_name (str): [description]

    Returns:
        A tuple of dictionaries (Tuple[Dict[str, str], Dict[str, str]]):
            nicknames_to_official
            official_to_short_name
    """
    nicknames_to_official: Dict[str, str] = {}
    official_to_short_name: Dict[str, str] = {}
    path = data_path(short_name)
    test_data_file(short_name)
    with open(path, "r", encoding="utf-8-sig") as file:
        csv_reader = csv.reader(file, delimiter=",")
        for row in csv_reader:
            official = row[0].strip()
            short = row[1].strip()
            nicknames_to_official[short.lower()] = official
            nicknames_to_official[official.lower()] = official
            official_to_short_name[official.lower()] = short
            for nickname in row[2:]:
                nicknames_to_official[nickname.lower()] = official
    # Green Party, with its different chapters, needs to be set by hand
    nicknames_to_official["green party"] = "Green Party"
    return nicknames_to_official, official_to_short_name


def read_csv_to_dataframe(
    short_name: str,
    separator: str = ","
) -> "pd.DataFrame":
    """Load csv into pandas dataframe.

    Args:
        short_name (str): Name of data file.
        separator (str, optional): Separator used in the file. Defaults to ",".

    Returns:
        data: dataframe
    """
    path = data_path(short_name)
    test_data_file(short_name)
    dataframe = pd.read_csv(path, sep=separator)
    return dataframe


PARTY_NICKNAMES, PARTY_SHORTNAMES = read_nicknames_file("party_nicknames.csv")
"""Official names from the nicknames we have on record."""

PARTY_COLORS = read_csv_to_dict("party_colors.csv")
"""Web color names for the parties."""

"""MIT License

Copyright (c) 2021 Stonehaven

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


r"""Function for searching for the current list of valid parties.

The colors themselves are from
https://en.wikipedia.org/wiki/Wikipedia:Index_of_United_Kingdom_political_parties_meta_attributes

Although there is no singular Green Party it is still likely
that someone will want to use our module to find a suitable color.
The historic color is supplied, along with a warning.
"""

from typing import Optional
import warnings
from . import data_tables
from . import exceptions


def by_official_party_name(party_name: Optional[str]) -> Optional[str]:
    """Return current hex color for a party.

    If no color is found for that party name try
    getting the official name using the politics.names submodule.

    Args:
        party_name (str): The official name of the party

    Raises:
        exceptions.NoColorForThisParty:
            If the given name is not in our data then an
            exception is raised.

    Returns:
        hexcode str: The (current) hex color for that party.
        For example:
            color("Brexit Party") -> "#12B6CF"
    """
    if party_name is None:
        return None

    if party_name in data_tables.PARTY_COLORS:
        if party_name == "Green Party":
            warnings.warn(("The Green Party has split into different chapters"
                           " in each country, each with different colors."
                           " The historic color (from when it was one party)"
                           " has been returned."), exceptions.UKPoliticsWarning)
        return data_tables.PARTY_COLORS[party_name]

    raise exceptions.NoColorForThisParty(party_name)

"""MIT License

Copyright (c) 2021 Stonehaven

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""